"""Test config for aiopegelonline."""
from __future__ import annotations

import json

import aiohttp
import pytest
from aioresponses import CallbackResult, aioresponses

from aiopegelonline import PegelOnline
from aiopegelonline.const import BASE_URL

from .const import MOCK_DATA, MOCK_STATION_DATA_DRESDEN


@pytest.fixture
def mock_aioresponse():
    """Mock a web request and provide a response."""
    with aioresponses() as m:
        yield m


@pytest.fixture
async def mock_pegelonline():
    """Return PegelOnline session."""
    session = aiohttp.ClientSession()
    api = PegelOnline(session)
    yield api
    await session.close()


@pytest.fixture
def mock_pegelonline_with_data(mock_aioresponse, mock_pegelonline):
    """Comfort fixture to initialize pegelonline session."""

    async def data_to_pegelonline() -> PegelOnline:
        """Initialize PegelOnline session."""
        for path, data in MOCK_DATA.items():
            mock_aioresponse.get(
                f"{BASE_URL}/{path}",
                status=data["status"],
                body=json.dumps(data["body"]),
                exception=data.get("exception"),
            )
        return mock_pegelonline

    return data_to_pegelonline


@pytest.fixture
def mock_pegelonline_with_cached_data(mock_aioresponse, mock_pegelonline):
    """Comfort fixture to initialize pegelonline session with cached data."""

    def cache_response(_, **kwargs) -> CallbackResult:
        etag = "etag_station_dresden"
        if (headers := kwargs.get("headers")) and headers.get("If-None-Match") == etag:
            return CallbackResult(status=304, body="", headers={"Etag": etag})
        return CallbackResult(
            status=200,
            body=json.dumps(MOCK_STATION_DATA_DRESDEN),
            headers={"Etag": etag},
        )

    async def data_to_pegelonline() -> PegelOnline:
        """Initialize PegelOnline session."""
        query = f"{BASE_URL}/stations/70272185-xxxx-xxxx-xxxx-43bea330dcae.json?prettyprint=false"
        mock_aioresponse.get(query, callback=cache_response, repeat=True)
        return mock_pegelonline

    return data_to_pegelonline



"""Constants for aiopegelonline tests."""

from aiohttp import ClientError

MOCK_STATION_DATA_DRESDEN: dict = {
    "uuid": "70272185-xxxx-xxxx-xxxx-43bea330dcae",
    "number": "501060",
    "shortname": "DRESDEN",
    "longname": "DRESDEN",
    "km": 55.63,
    "agency": "STANDORT DRESDEN",
    "longitude": 13.738831783620384,
    "latitude": 51.054459765598125,
    "water": {"shortname": "ELBE", "longname": "ELBE"},
}

MOCK_MEASUREMENT_DATA_HANAU_BRIDGE: dict = {
    "uuid": "07374faf-xxxx-xxxx-xxxx-adc0e0784c4b",
    "number": "24700347",
    "shortname": "HANAU BRÜCKE DFH",
    "longname": "HANAU BRÜCKE DFH",
    "km": 56.398,
    "agency": "ASCHAFFENBURG",
    "water": {"shortname": "MAIN", "longname": "MAIN"},
    "timeseries": [
        {
            "shortname": "DFH",
            "longname": "DURCHFAHRTSHÖHE",
            "unit": "cm",
            "equidistance": 15,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:45:00+02:00",
                "value": 715,
            },
            "gaugeZero": {
                "unit": "m. ü. NHN",
                "value": 106.501,
                "validFrom": "2019-11-01",
            },
        }
    ],
}

MOCK_STATION_DATA_WUERZBURG: dict = {
    "uuid": "915d76e1-xxxx-xxxx-xxxx-4d144cd771cc",
    "number": "24300600",
    "shortname": "WÜRZBURG",
    "longname": "WÜRZBURG",
    "km": 251.97,
    "agency": "SCHWEINFURT",
    "longitude": 9.925968763247354,
    "latitude": 49.79620901036012,
    "water": {"shortname": "MAIN", "longname": "MAIN"},
}

MOCK_MEASUREMENT_DATA_WUERZBURG: dict = {
    **MOCK_STATION_DATA_WUERZBURG,
    "timeseries": [
        {
            "shortname": "W",
            "longname": "WASSERSTAND ROHDATEN",
            "unit": "cm",
            "equidistance": 15,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:15:00+02:00",
                "value": 159,
                "stateMnwMhw": "normal",
                "stateNswHsw": "normal",
            },
            "gaugeZero": {
                "unit": "m. ü. NHN",
                "value": 164.511,
                "validFrom": "2019-11-01",
            },
        },
        {
            "shortname": "LT",
            "longname": "LUFTTEMPERATUR",
            "unit": "°C",
            "equidistance": 60,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:00:00+02:00",
                "value": 21.2,
            },
        },
        {
            "shortname": "WT",
            "longname": "WASSERTEMPERATUR",
            "unit": "°C",
            "equidistance": 60,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:00:00+02:00",
                "value": 22.1,
            },
        },
        {
            "shortname": "VA",
            "longname": "FLIESSGESCHWINDIGKEIT",
            "unit": "m/s",
            "equidistance": 15,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:15:00+02:00",
                "value": 0.58,
            },
        },
        {
            "shortname": "O2",
            "longname": "SAUERSTOFFGEHALT",
            "unit": "mg/l",
            "equidistance": 60,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:00:00+02:00",
                "value": 8.4,
            },
        },
        {
            "shortname": "PH",
            "longname": "PH-WERT",
            "unit": "--",
            "equidistance": 60,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:00:00+02:00",
                "value": 8.1,
            },
        },
        {
            "shortname": "Q",
            "longname": "ABFLUSS",
            "unit": "m³/s",
            "equidistance": 15,
            "currentMeasurement": {
                "timestamp": "2023-07-26T19:00:00+02:00",
                "value": 102,
            },
        },
    ],
}

MOCK_DATA: dict = {
    "stations.json?prettyprint=false": {
        "status": 200,
        "body": [
            MOCK_STATION_DATA_DRESDEN,
            MOCK_STATION_DATA_WUERZBURG,
        ],
    },
    "stations.json?prettyprint=false&latitude=13&longitude=51&radius=25": {
        "status": 200,
        "body": [
            MOCK_STATION_DATA_DRESDEN,
        ],
    },
    "stations.json?prettyprint=false&latitude=10&longitude=45&radius=25": {
        "status": 200,
        "body": [],
    },
    "stations/70272185-xxxx-xxxx-xxxx-43bea330dcae.json?prettyprint=false": {
        "status": 200,
        "body": MOCK_STATION_DATA_DRESDEN,
    },
    "stations/INVALID_UUID.json?prettyprint=false": {
        "status": 404,
        "body": {
            "status": 404,
            "message": "station not found",
        },
    },
    "stations/CONNECT_ERROR.json?prettyprint=false": {
        "status": None,
        "body": None,
        "exception": ClientError,
    },
    "stations/915d76e1-xxxx-xxxx-xxxx-4d144cd771cc.json?prettyprint=false&includeTimeseries=true&includeCurrentMeasurement=true": {
        "status": 200,
        "body": MOCK_MEASUREMENT_DATA_WUERZBURG,
    },
    "stations/07374faf-xxxx-xxxx-xxxx-adc0e0784c4b.json?prettyprint=false&includeTimeseries=true&includeCurrentMeasurement=true": {
        "status": 200,
        "body": MOCK_MEASUREMENT_DATA_HANAU_BRIDGE,
    },
}



"""aiopegelonline models."""

from __future__ import annotations

from dataclasses import dataclass


class Station:
    """Representation of a station."""

    def __init__(self, data: dict) -> None:
        """Initialize station class."""
        self.uuid: str = data["uuid"]
        self.name: str = data["longname"]
        self.agency: str = data["agency"]
        self.river_kilometer: float | None = data.get("km")
        self.longitude: float | None = data.get("longitude")
        self.latitude: float | None = data.get("latitude")
        self.water_name: str = data["water"]["longname"]
        self.base_data_url: str = (
            f"https://www.pegelonline.wsv.de/gast/stammdaten?pegelnr={data['number']}"
        )

    def as_dict(self) -> dict[str, str | float | None]:
        """Return data das dict."""
        return {
            "uuid": self.uuid,
            "name": self.name,
            "agency": self.agency,
            "river_kilometer": self.river_kilometer,
            "longitude": self.longitude,
            "latitude": self.latitude,
            "water_name": self.water_name,
            "base_data_url": self.base_data_url,
        }


class CurrentMeasurement:
    """Representation of a current measurement."""

    def __init__(self, data: dict) -> None:
        """Initialize current measurment class."""
        self.uom: str = data["unit"]
        self.value: float = data["currentMeasurement"]["value"]

    def as_dict(self) -> dict[str, str | float]:
        """Return data das dict."""
        return {
            "uom": self.uom,
            "value": self.value,
        }


class StationMeasurements:
    """Representation of station measurements."""

    def __init__(self, data: list[dict]) -> None:
        """Initialize station measurments class."""
        self.air_temperature: CurrentMeasurement | None = None
        self.clearance_height: CurrentMeasurement | None = None
        self.oxygen_level: CurrentMeasurement | None = None
        self.ph_value: CurrentMeasurement | None = None
        self.water_speed: CurrentMeasurement | None = None
        self.water_flow: CurrentMeasurement | None = None
        self.water_level: CurrentMeasurement | None = None
        self.water_temperature: CurrentMeasurement | None = None

        for measurement in data:
            if measurement["shortname"] == "DFH":
                self.clearance_height = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "LT":
                self.air_temperature = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "O2":
                self.oxygen_level = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "PH":
                self.ph_value = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "Q":
                self.water_flow = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "VA":
                self.water_speed = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "W":
                self.water_level = CurrentMeasurement(measurement)
            elif measurement["shortname"] == "WT":
                self.water_temperature = CurrentMeasurement(measurement)

    def as_dict(self) -> dict[str, CurrentMeasurement | None]:
        """Return data das dict."""
        return {
            "air_temperature": self.air_temperature,
            "clearance_height": self.clearance_height,
            "oxygen_level": self.oxygen_level,
            "ph_value": self.ph_value,
            "water_speed": self.water_speed,
            "water_flow": self.water_flow,
            "water_level": self.water_level,
            "water_temperature": self.water_temperature,
        }


@dataclass
class CacheEntry:
    """Representation of response cache entry."""

    etag: str | None
    result: dict | None



"""aiopegelonline exceptions."""
from __future__ import annotations


class PegelonlineError(Exception):
    """Base class for pegelonline errors."""


class PegelonlineDataError(PegelonlineError):
    """Raised to indicate invalid data."""

    def __init__(self, code: int, message: str = ""):
        """Initialize JSON RPC errors."""
        self.code = code
        self.message = message
        super().__init__(code, message)



"""Constants for aiopegelonline."""
import asyncio
import logging

import aiohttp

LOGGER = logging.getLogger(__package__)

BASE_URL = "https://www.pegelonline.wsv.de/webservices/rest-api/v2"

CONNECT_ERRORS = (
    aiohttp.ClientError,
    asyncio.TimeoutError,
    OSError,
)



"""Pegelonline library."""

from __future__ import annotations

import json

from aiohttp.client import ClientSession

from .const import BASE_URL, CONNECT_ERRORS, LOGGER
from .exceptions import PegelonlineDataError
from .models import CacheEntry, Station, StationMeasurements


class PegelOnline:
    """Pegelonline api."""

    def __init__(self, aiohttp_session: ClientSession) -> None:
        """Pegelonline api init."""
        self.session: ClientSession = aiohttp_session
        self.cache: dict[str, CacheEntry] = {}

    async def _async_do_request(self, url: str, params: dict):
        """Perform an async request."""
        cache_key = f"{url}_{params}"
        if cache_key not in self.cache:
            self.cache[cache_key] = CacheEntry(None, None)

        cache_entry = self.cache[cache_key]

        headers = {}
        if (etag := cache_entry.etag) is not None:
            headers = {"If-None-Match": etag}

        LOGGER.debug("REQUEST url: %s params: %s headers: %s", url, params, headers)
        try:
            async with self.session.get(url, params=params, headers=headers) as resp:
                result = await resp.text()
                LOGGER.debug("RESPONSE status: %s text: %s", resp.status, result)

                if resp.status == 304:  # 304 = not modified
                    return cache_entry.result

                if cache_entry.etag is None:
                    cache_entry.etag = resp.headers.get("Etag")

                result = json.loads(result)
                cache_entry.result = result
                if resp.status != 200:
                    raise PegelonlineDataError(
                        result.get("status"), result.get("message")
                    )
        except CONNECT_ERRORS as err:
            LOGGER.debug("connection error", exc_info=True)
            LOGGER.error(
                "Error while getting data: %s: %s",
                err.__class__.__name__,
                err.__class__.__cause__,
            )
            raise err

        return result

    async def async_get_all_stations(self) -> dict[str, Station]:
        """Get all stations."""
        stations = await self._async_do_request(
            f"{BASE_URL}/stations.json", {"prettyprint": "false"}
        )

        result = {}
        for station in stations:
            result[station["uuid"]] = Station(station)

        LOGGER.debug("all stations: %s", result)

        return result

    async def async_get_nearby_stations(
        self, latitude: float, longitude: float, radius: int
    ) -> dict[str, Station]:
        """Get stations within defined radius at given position."""
        stations = await self._async_do_request(
            f"{BASE_URL}/stations.json",
            {
                "prettyprint": "false",
                "latitude": latitude,
                "longitude": longitude,
                "radius": radius,
            },
        )

        result = {}
        for station in stations:
            result[station["uuid"]] = Station(station)

        LOGGER.debug("nearby stations: %s", result)

        return result

    async def async_get_station_details(self, uuid: str) -> Station:
        """Get station details."""
        station = await self._async_do_request(
            f"{BASE_URL}/stations/{uuid}.json", {"prettyprint": "false"}
        )

        result = Station(station)

        LOGGER.debug("station %s details: %s", uuid, result)

        return result

    async def async_get_station_measurements(self, uuid: str) -> StationMeasurements:
        """Get all current measurements of a station."""
        station = await self._async_do_request(
            f"{BASE_URL}/stations/{uuid}.json",
            {
                "prettyprint": "false",
                "includeTimeseries": "true",
                "includeCurrentMeasurement": "true",
            },
        )

        result = StationMeasurements(station["timeseries"])

        LOGGER.debug("station %s measurements: %s", uuid, result)

        return result



"""Tests for aiopegelonline."""
from __future__ import annotations

import pytest
from aiohttp import ClientError

from aiopegelonline.exceptions import PegelonlineDataError
from aiopegelonline.models import Station, StationMeasurements


async def test_get_all_stations(mock_pegelonline_with_data):
    """Test async_get_all_stations."""
    api = await mock_pegelonline_with_data()
    stations = await api.async_get_all_stations()
    assert len(stations) == 2

    station = stations["70272185-xxxx-xxxx-xxxx-43bea330dcae"]
    assert isinstance(station, Station)
    assert station.as_dict() == {
        "uuid": "70272185-xxxx-xxxx-xxxx-43bea330dcae",
        "name": "DRESDEN",
        "agency": "STANDORT DRESDEN",
        "river_kilometer": 55.63,
        "longitude": 13.738831783620384,
        "latitude": 51.054459765598125,
        "water_name": "ELBE",
        "base_data_url": "https://www.pegelonline.wsv.de/gast/stammdaten?pegelnr=501060",
    }

    station = stations["915d76e1-xxxx-xxxx-xxxx-4d144cd771cc"]
    assert isinstance(station, Station)
    assert station.uuid == "915d76e1-xxxx-xxxx-xxxx-4d144cd771cc"
    assert station.name == "WÜRZBURG"
    assert station.agency == "SCHWEINFURT"
    assert station.river_kilometer == 251.97
    assert station.longitude == 9.925968763247354
    assert station.latitude == 49.79620901036012
    assert station.water_name == "MAIN"
    assert (
        station.base_data_url
        == "https://www.pegelonline.wsv.de/gast/stammdaten?pegelnr=24300600"
    )


async def test_get_nearby_stations(mock_pegelonline_with_data):
    """Test async_get_nearby_stations."""
    api = await mock_pegelonline_with_data()
    stations = await api.async_get_nearby_stations(13, 51, 25)
    assert len(stations) == 1

    station = stations["70272185-xxxx-xxxx-xxxx-43bea330dcae"]
    assert isinstance(station, Station)
    assert station.uuid == "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    assert station.name == "DRESDEN"
    assert station.agency == "STANDORT DRESDEN"
    assert station.river_kilometer == 55.63
    assert station.longitude == 13.738831783620384
    assert station.latitude == 51.054459765598125
    assert station.water_name == "ELBE"
    assert (
        station.base_data_url
        == "https://www.pegelonline.wsv.de/gast/stammdaten?pegelnr=501060"
    )


async def test_get_nearby_stations_no_stations(mock_pegelonline_with_data):
    """Test async_get_nearby_stations."""
    api = await mock_pegelonline_with_data()
    stations = await api.async_get_nearby_stations(10, 45, 25)
    assert len(stations) == 0


async def test_get_station_details(mock_pegelonline_with_data):
    """Test async_get_station_details."""
    api = await mock_pegelonline_with_data()
    station = await api.async_get_station_details(
        "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    )
    assert isinstance(station, Station)
    assert station.uuid == "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    assert station.name == "DRESDEN"
    assert station.agency == "STANDORT DRESDEN"
    assert station.river_kilometer == 55.63
    assert station.longitude == 13.738831783620384
    assert station.latitude == 51.054459765598125
    assert station.water_name == "ELBE"
    assert (
        station.base_data_url
        == "https://www.pegelonline.wsv.de/gast/stammdaten?pegelnr=501060"
    )


async def test_get_station_details_invalid(mock_pegelonline_with_data):
    """Test async_get_station_details with invalid uuid."""
    api = await mock_pegelonline_with_data()
    with pytest.raises(PegelonlineDataError):
        await api.async_get_station_details("INVALID_UUID")


async def test_get_station_details_connection_error(mock_pegelonline_with_data):
    """Test async_get_station_details with connection error."""
    api = await mock_pegelonline_with_data()
    with pytest.raises(ClientError):
        await api.async_get_station_details("CONNECT_ERROR")


async def test_get_station_measurements(mock_pegelonline_with_data):
    """Test async_get_station_measurements."""
    api = await mock_pegelonline_with_data()
    measurement = await api.async_get_station_measurements(
        "915d76e1-xxxx-xxxx-xxxx-4d144cd771cc"
    )
    assert isinstance(measurement, StationMeasurements)
    assert measurement.air_temperature is not None
    assert measurement.air_temperature.uom == "°C"
    assert measurement.air_temperature.value == 21.2
    assert measurement.clearance_height is None
    assert measurement.oxygen_level is not None
    assert measurement.oxygen_level.uom == "mg/l"
    assert measurement.oxygen_level.value == 8.4
    assert measurement.ph_value is not None
    assert measurement.ph_value.uom == "--"
    assert measurement.ph_value.value == 8.1
    assert measurement.water_speed is not None
    assert measurement.water_speed.uom == "m/s"
    assert measurement.water_speed.value == 0.58
    assert measurement.water_flow is not None
    assert measurement.water_flow.uom == "m³/s"
    assert measurement.water_flow.value == 102
    assert measurement.water_level is not None
    assert measurement.water_level.uom == "cm"
    assert measurement.water_level.value == 159
    assert measurement.water_temperature is not None
    assert measurement.water_temperature.uom == "°C"
    assert measurement.water_temperature.value == 22.1

    measurement = await api.async_get_station_measurements(
        "07374faf-xxxx-xxxx-xxxx-adc0e0784c4b"
    )
    assert isinstance(measurement, StationMeasurements)
    assert measurement.clearance_height is not None
    assert measurement.clearance_height.uom == "cm"
    assert measurement.clearance_height.value == 715
    assert measurement.clearance_height.as_dict() == {
        "uom": "cm",
        "value": 715,
    }
    assert measurement.as_dict() == {
        "air_temperature": None,
        "clearance_height": measurement.clearance_height,
        "oxygen_level": None,
        "ph_value": None,
        "water_speed": None,
        "water_flow": None,
        "water_level": None,
        "water_temperature": None,
    }


async def test_get_station_details_cached(mock_pegelonline_with_cached_data):
    """Test response cache."""
    api = await mock_pegelonline_with_cached_data()
    station = await api.async_get_station_details(
        "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    )
    assert isinstance(station, Station)
    assert station.uuid == "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    assert station.name == "DRESDEN"
    assert station.agency == "STANDORT DRESDEN"
    assert station.river_kilometer == 55.63
    assert station.longitude == 13.738831783620384
    assert station.latitude == 51.054459765598125
    assert station.water_name == "ELBE"

    station = await api.async_get_station_details(
        "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    )
    assert isinstance(station, Station)
    assert station.uuid == "70272185-xxxx-xxxx-xxxx-43bea330dcae"
    assert station.name == "DRESDEN"
    assert station.agency == "STANDORT DRESDEN"
    assert station.river_kilometer == 55.63
    assert station.longitude == 13.738831783620384
    assert station.latitude == 51.054459765598125
    assert station.water_name == "ELBE"
